 /*****************************************************************************
 *                                                                            *
 *   Copyright (C) 2004-2006 by Michael Schulze                               *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef IPOD_H
#define IPOD_H

#include <qfile.h>
#include <qstring.h>

#include "itunesdb.h"

#include "ipoddevicedetails.h"
#include "ipodsysinfo.h"

#define LOGFILEPREFIX "/kio_ipod-"
#define LOG_DEFAULT true

/**
This class represents all functionalities to access and modify information about the iPod/itunesDB

@author Michael Schulze
*/
class IPod {
public:

    /**
     * Constructs a new IPod instance for an iPod at the given path. Does not actually
     * read the contents - you need to open it with @c open().
     */
    IPod( const QString& ipodBase );

    /**
     * The destructor.
     */
    virtual ~IPod();

    /**
     * Tries to open an ipod. That means reading all important information from the device.
     * @return true if successful, otherwise false
     */
    bool open();

    /**
     * initializes this ipod by creating the nessessary directories and files
     * This IPod instance then contains the information about the
     * initialized ipod - no need to call open() afterwards.
     */
    void initialize( const QString& title );

    /**
     * Returns true if open() was called successfully for this instance.
     * @return true if open() was called successfully for this instance.
     */
    bool isOpen();

    /**
     * Returns true if the ipod represented by this instance is still connected to the system
     * @return true if the ipod represented by this instance is still connected to the system false otherwise
     */
    bool isStillConnected();

    /**
     * Clears all data we got from this instance whether or not it was changed.
     * Resets the state of this instance.
     */
    void close();

    /**
     * Returns the name of the iPod
     */
    QString getName() const;

    /**
     * Sets the name of the iPod. When the iPod is not opened (isOpen() == false) nothing is changed
     */
    void setName(const QString& name);

    /**
     * returns the error occured during read or write
     */
    const QString& getItunesDBError() const;

    /**
     * Returns the IPodSysInfo for this IPod instance (const).
     * @return the IPodSysInfo for this IPod instance.
     */
    const IPodSysInfo& getSysInfo() const;

    /**
     * Returns the IPodSysInfo for this IPod instance.
     * @return the IPodSysInfo for this IPod instance.
     */
    IPodSysInfo& getSysInfo();

    /**
     * Returns the ITunesDB for this iPod (const).
     * @return the ITunesDB for this iPod.
     */
    const ITunesDB& getITunesDB() const;

    /**
     * Returns the ITunesDB for this iPod.
     * @return the ITunesDB for this iPod.
     */
    ITunesDB& getITunesDB();

    /**
     * Writes back all the changed data to the device. This only includes
     * control structures like the iTunesDB and DeviceInfo data.
     */
    void synchronize();

    /**
     * Returns true if this instances data differ somehow from the original
     * data from the iPod and need to be synchronized.
     */
    bool isDirty() const { return m_itunesdb.isDirty(); }

    /**
     * Lock control functions to prevent concurrent access to the iPod from different programs
     */
    void lock(bool write_lock);
    bool isLocked();
    void unlock();

    const QString& getBasePath() const { return m_ipodBase; }

    static QString createDistinctIPodName(const IPod& ipod);

    // This method can be removed when we're able to handle podcasts
    bool hasPodcasts() { return m_itunesdb.hasPodcasts(); }

protected:
    QString m_ipodBase;

private:
    IPodSysInfo m_sysInfo;
    ITunesDB m_itunesdb;
    IPodDeviceDetails m_deviceDetails;

};

#endif
