 /*****************************************************************************
 *   Copyright (C) 2004 by Andrew de Quincey                                  *
 *   adq@lidskialf.net                                                        *
 *   Copyright (C) 2005 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef IPODSYSINFO_H
#define IPODSYSINFO_H

#include <qobject.h>
#include <qmap.h>
#include <qstring.h>
#include <qdir.h>

#define IPOD_CONTROLDIR "/iPod_Control"
#define IPOD_MUSICDIR   "Music"

/**
 * @author Andrew de Quincey, Michael Schulze
 */
class IPodSysInfo {
public:
    IPodSysInfo( const QString& filename );
    ~IPodSysInfo();

    /**
     * Loads the /Device/SysInfo file and figures out where the music directories are.
     */
    bool load();

    void refreshDiskUsageStats();

    /**
    * Get the keys to all known system information properties.
    *
    * @return Key list.
    */
    QValueList<QString> getKeys() const;

    /**
    * Retrieve an individual property by key
    *
    * @param key The key to look up.
    * @return The value, or QString::null if unknown.
    */
    QString getValue(const QString& key) const;

    /**
     * Returns the total disk space available on the device (filesystem wise)
     * in kBytes
     */
    unsigned long getTotalDiskSpaceKB();

    /**
     * Returns the currently available disk space on the device (filesystem wise)
     * in kBytes
     */
    unsigned long getAvailableDiskSpaceKB();

    /**
     * Returns the currently used disk space on the device (filesystem wise)
     * in kBytes
     */
    unsigned long getUsedDiskSpaceKB();

    /**
     * Returns the number of fXX directories inside the "Music" dir
     */
    int getNumTrackFileDirs() const { return m_numTrackFileDirs; }

    /**
     * Returns the directory where the music files reside.
     */
    const QDir& getMusicDir() const { return m_musicDir; }

    // const QString& getDeviceName() const { return m_deviceName; }

    /**
     * Returns the device mountpoint (or drive letter on windows)
     */
    const QString& getDeviceMountpoint() const { return m_ipodBase; }

    /**
     * clears all information from this sysinfo instance.
     */
    void clear();

    /**
     * Returns the path to the "iPod_Control" directory on the iPod.
     * @return the path to the "iPod_Control" directory on the iPod.
     */
    QString getIPodControlDirName() { return QString( m_ipodBase + IPOD_CONTROLDIR ); }

    /**
     * Creates the "Music" directory and the fXX subdirs.
     * @return true if successful, false if such a directory already exists
     */
    bool createMusicDirs();

    /**
     * finds the first track filename for the given trackid.
     * the track filename will be formed like<br>
     * kpod[decimal encoded ID]<br>
     * the file extension needs to be added later
     */
    QString findNewTrackFileName( Q_UINT32 id );

private:

    QDir findMusicDir( const QString& ipodbase, bool * found = NULL ) const;

    void ensureDiskUsageStats();
    QMap<QString, QString> m_details;
    QString m_ipodBase;
    // QString m_deviceName;
    QDir m_musicDir;
    uint m_numTrackFileDirs;

    unsigned long m_kbytesTotal;
    unsigned long m_kbytesAvail;
};

#endif
