 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef ITUNESDB_H
#define ITUNESDB_H

#include <qmap.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qptrlist.h>
#include <qstringlist.h>

#include <vector>

#include "playlist.h"
#include "playlistitem.h"
#include "track.h"
#include "ipodsysinfo.h"

#define RECENTLY_PLAYED_LIST_NAME "KPOD:Recently Played"

/**
@author Michael Schulze
*/

class ITunesDBTrack : public itunesdb::Track, itunesdb::utils::NonCopyAble {

    friend class ITunesDB;

    class ITunesDB& m_itunesdb;

    ITunesDBTrack( Q_UINT32 trackid, ITunesDB& itunesdb );
    ITunesDBTrack( const ITunesDBTrack& src );

public:
    virtual ~ITunesDBTrack();

    void setDateAdded( Q_UINT32 date );

    /**
     * Returns the file on the iPod this Track instance refers to.
     */
    QString getFilePath() const;

protected:
    virtual void setDirty( bool dirtyFlag = true );
};


/**
 * This class represents all the rules and limits making up a smart playlist.
 * see @c itunesdb::SmartPlaylistRuleSet for further information.
 */
class ITunesDBSPLRuleSet
    : public itunesdb::SmartPlaylistRuleSet,
      public itunesdb::TrackPredicate,
      itunesdb::utils::NonCopyAble
{

    const class ITunesDB& mItunesdb;
    class ITunesDBPlaylist& mPlaylist;

    itunesdb::TrackComparatorPtr getComparatorFor( itunesdb::SPLLimitSort limitSort ) const;

    void applyLimits( itunesdb::TrackPtrList& trackList ) const;

    uint mLastSeenItunesDBVersion;
    uint mLastSeenRuleSetVersion;
    uint mRuleSetVersion;

protected:

    virtual itunesdb::SmartPlaylistRule * createStringRule(
        itunesdb::SPLStringField field,
        itunesdb::SPLStringAction action,
        const QString& testString );

    virtual itunesdb::SmartPlaylistRule * createUIntRule(
        itunesdb::SPLUIntField field, itunesdb::SPLUIntAction action,
        Q_UINT64 fromValue, Q_UINT64 toValue,
        Q_INT64  fromDate,  Q_INT64  toDate,
        Q_UINT64 fromUnits, Q_UINT64 toUnits );

    virtual bool operator() ( const itunesdb::Track& track );

public:

    ITunesDBSPLRuleSet( const itunesdb::SmartPlaylistRuleSet& src,
                        ITunesDBPlaylist& playlist, const ITunesDB& itunesdb );

    ITunesDBSPLRuleSet( ITunesDBPlaylist& playlist, const ITunesDB& itunesdb );

    virtual ~ITunesDBSPLRuleSet();

    /**
     * adds a string based rule to the rule set and returns true if successful.
     */
    virtual bool addStringRule( itunesdb::SPLStringField field,
                 itunesdb::SPLStringAction action, const QString& testString, int position = -1 );

    /**
     * adds a uint based rule to the rule set and returns true if successful.
     */
    virtual bool addUIntRule( itunesdb::SPLUIntField field,
                              itunesdb::SPLUIntAction action,
                              Q_UINT64 fromValue, Q_UINT64 toValue,
                               Q_INT64 fromDate,   Q_INT64 toDate,
                              Q_UINT64 fromUnits, Q_UINT64 toUnits,
                              int position = -1 );

    /**
     * Refreshes the playlist associated with this set of rules. If the database didn't
     * change since the last refresh and @c force is false the playlist won't be updated.
     * @param force forces the rule set to update the playlist even if the database didn't change.
     */
    void refreshPlaylist( bool force = false );

};


/**
 * Special playlist implementation to use with the ITunesDB class. This adds important
 * functionalities like setting the sort order and updating the contents of smart playlists.<br>
 * To create a new ITunesDBPlaylist instance use the
 * @c ITunesDB::createNewPlaylist( const QString& ) factory method and use the
 * @c ITunesDB::addPlaylist( ITunesDBPlaylist* ) to add it to the database.<br>
 * Here is an example on how to create a "recently played" Smart Playlist:
 * <pre><code>
 *  // create a new playlist
 *  ITunesDBPlaylist * recentlyPlayed = itunesdb.createNewPlaylist( "recently played" );
 *  recentlyPlayed->setSortOrder( Playlist::SORTORDER_TIME_PLAYED );
 *
 *  // make it a smart playlist and get the ruleset
 *  SmartPlaylistRuleSet& ruleSet = recentlyPlayed->enableSmartPlaylist();
 *
 *  // add the rule "last played in the last 2 weeks"
 *  ruleSet.addInTheLastRule( FIELD_LASTPLAYED, false, -2, IN_THE_LAST_WEEKS );
 *
 *  // limit to the first 30 songs, sorted by last played date
 *  ruleSet.setLimits( LIMIT_SORT_LASTPLAYED, LIMIT_TYPE_SONGS, 30 );
 *
 *  // match ANY rule (doesn't really matter since we only have one rule)
 *  ruleSet.setMatchAnyFlag( true );
 *
 *  // enable live update (updates the playlist on the go)
 *  ruleSet.setLiveUpdateFlag( true );
 *
 *  // finally add the playlist to the database
 *  itunesdb.addPlaylist( recentlyPlayed );
 * </code></pre>
 */
class ITunesDBPlaylist : public itunesdb::Playlist {

    friend class ITunesDB;

    ITunesDB& m_itunesdb;
    ITunesDBPlaylist( const ITunesDBPlaylist& );
    ITunesDBPlaylist( ITunesDB& itunesdb );
    ITunesDBPlaylist( ITunesDB& itunesdb, const itunesdb::Playlist& playlist );
    itunesdb::TrackComparatorPtr getTrackComparatorFor( Q_UINT32 sortOrder );

    /**
     * @see @c itunesdb::ListItem
     */
    void doPostParsingChecks();

public:

    /**
     * An entry in this playlist. It directly refers to an existing track in the
     * database.
     */
    class Entry : public itunesdb::PlaylistItem {
        uint m_mhitTrackPos;
    public:
        Entry( Q_UINT32 id, uint trackpos );
        virtual ~Entry();
        void setMHITTrackPos( uint trackpos );
        uint getMHITTrackPos();
    };

    virtual ~ITunesDBPlaylist();

    /**
     * Creates a new instance of the Entry class.
     */
    virtual itunesdb::PlaylistItem * createNewItem( Q_UINT32 itemid );

    /**
     * Updates the content of this smart playlist against the database
     * Nothing happens if this playlist is not a smart playlist.
     * @see itunesdb::Playlist::enableSmartPlaylist()
     * @see itunesdb::Playlist::getSmartPlaylistRules()
     */
    void updateSmartPlaylist( bool force = false );

    /**
     * Sets the sort order field.
     * @see itunesdb::Playlist::ItdbPlaylistSortOrder
     */
    void setSortOrder( Sortorder sortOrder );

    /**
     * Returns the track the playlist entry at the given position is referring to
     */
    itunesdb::Track * getTrackAt( uint position );

protected:

    /**
     * reads mhods with type above type 15 (mhods not containing strings)
     */
    virtual void readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen );

    virtual void setDirty( bool dirtyFlag = true );

    virtual itunesdb::SmartPlaylistRuleSet * createNewSplRuleSet();

};


/**
 * Class representing the contents of your iPods ITunesDB file
 */
class ITunesDB : itunesdb::utils::NonCopyAble {
    class ItunesDBReaderAdapter;    // Adapter class for reading the itunesdb file
    class ItunesDBWriterAdapter;    // Adapter class for writing the itunesdb file

    // friends
    friend class ITunesDBPlaylist;
    friend class ITunesDBSPLRuleSet;
    friend class ITunesDBTrack;

    typedef QMap<Q_UINT32,uint> TrackIDMap;
    typedef std::vector<ITunesDBTrack *> TrackList;

    bool mDoDirtyFlagHandling;
    uint mInternalVersion;

    // FIXME implement shared PImpl thus making it copyable

public:

    /**
     * An Iterator over a set of tracks
     */
    class TrackIterator {
    public:
        TrackIterator( const TrackList& trackList, itunesdb::TrackPredicate * predicate);
        ~TrackIterator();

        bool hasNext();
        ITunesDBTrack * next();
    private:
        void calcNext();
        TrackList::const_iterator m_pos;
        TrackList::const_iterator m_end;
        TrackList::const_iterator m_next;
        itunesdb::TrackPredicate * m_predicate;
    };

    QString error;

    ITunesDB( IPodSysInfo& ipodSysInfo );

    virtual ~ITunesDB();

    /**
     * Opens the itunesdb file at the given ipod mountpoint
     * @return true if the itunesdb does exist and could be opened. For parse errors check the error member afterwards
     */
    bool open();

    /**
     * Returns true if open() has already been successfully called.
     * @return true if open has already been successfully called, false otherwise
     */
    bool isOpen();

    /**
     * Writes the (changed) database back to the ipod
     * @param filename the name of the file to be written. If no filename is given the database will be written to the file read whith open()
     * @return true if writing the database was successful
     */
    bool writeDatabase(const QString& filename = QString());

    /**
     * Returns true if the itunesdb file has been changed thus needs to be reloaded or does not exist anymore
     */
    bool dbFileChanged() const;

    /**
     * Returns the name of the filename given with the open() call
     */
    QString getFilename() const { return itunesdbfile.name(); }

    /**
     * Creates a new empty playlist
     */
    ITunesDBPlaylist * createNewPlaylist( const QString& title );

    /**
     * Creates a new empty track
     */
    ITunesDBTrack * createEmptyTrack( Q_UINT32 trackid );

    /**
     * Returns a new itunesdb::Track instance for adding a new track to the database.
     * The new instance gets initialized with an unused trackid and trackpath.
     * Subsequent calls will get itunesdb::Track with the same trackid until the returned track is as added
     * via addTrack() thus a call to addTrack() has to be done afterwards.
     * @param fileExtension file extension of the file representing the track. If you don't know the file extension yet leave empty and set it with Track::setFileExtension() later
     */
    ITunesDBTrack * createNewTrack( const QString& fileExtension = QString::null );

    /**
     * Adds a new track to the collection. This only adds the control information
     * to the database, the track itself (the file) has to by copied by the user.
     * @param track the track to add to the database
     * @return true if the track was successfully added
     */
    bool addTrack( ITunesDBTrack * track );

    /**
     * returns the Track corresponding to the given ID
     * @param id ID of the track
     * @return the Track corresponding to the given ID
     */
    ITunesDBTrack * getTrackByID( const Q_UINT32 id ) const;

    /**
     * Returns the first track found with the given dbid
     */
    ITunesDBTrack * findTrackByDBID( const Q_UINT64 dbid ) const;

    /**
     * Returns the MHIT position of the track with the given ID or -1 if there's no such track.
     * @param id ID of the track
     * @return the position of the track in the MHIT section
     */
    int getMHITTrackPosFor( Q_UINT32 id ) const;

    /**
     * returns the Track at the specified position in the itunesdb
     * @param mhitTrackPos the position of the desired track
     * @return the track at the specified position or NULL if the given position is out of range
     */
    ITunesDBTrack * getTrackAt( uint mhitTrackPos) const;

    /**
     * Returns the first Track where the given predicate returned true
     * The given trackpredicate needs to contain a method that gets a track, and returns a bool
     * if it's the desired track
     * @return the first Track where the given predicate returned true
     * @see class TrackPredicate
     */
    ITunesDBTrack * findFirstTrackBy( itunesdb::TrackPredicate& predicate ) const;


    /**
     * Returns all the Tracks where the given @c itunesdb::TrackPredicate returned true
     * Be aware that the returned TrackIterator depends on the given predicate, so do not
     * delete it as long as you need access to the Iterator
     * @see @c itunesdb::TrackPredicates for a list of predefined predicates
     */
    TrackIterator getTracksBy( itunesdb::TrackPredicate * predicate ) const;


    /**
     * Returns all the Tracks where the given @c itunesdb::TrackPredicate returned true
     * @see @c itunesdb::TrackPredicates for a list of predefined predicates
     */
    itunesdb::TrackPtrList * getTracksBy( itunesdb::TrackPredicate& predicate, itunesdb::TrackPtrList& buffer ) const;

    /**
     * returns the Track found by the given information or NULL if no such Track could be found
     * @param artistname the name of the artist
     * @param albumname the name of the album
     * @param title the title of the track
     * @param trackNum the position of the track in its album
     * @return the Track found or NULL if no track matched the given criteria.
     */
    ITunesDBTrack * findTrack(const QString& artistname, const QString& albumname, const QString& title, Q_UINT32 trackNum = 0 ) const;

    /**
     * Fills the given QStringList with the names of all artists in the database
     * @return a pointer to the StringList
     */
    QStringList * getArtists( QStringList& buffer ) const;

    /**
     * Fills the given QStringList with the names of all albums in the database
     */
    QStringList * getAllAlbums( QStringList& buffer ) const;

    /**
     * Fills the given QStringList with the names of the albums by the given artist
     * @param artistname the name of the artist to look for
     * @param buffer the list to put the names of the albums found
     * @return a pointer to the given QStringList or NULL if no such artist exists
     */
    QStringList * getAlbumsByArtist( const QString& artistname, QStringList& buffer ) const;

    /**
     * Returns true if the given artist exists.
     * @param artistname the artists name to check for.
     * @return true if the artist exists, false otherwise.
     */
    bool containsArtist( const QString& artistname ) const;

    /**
     * Returns true if the given album exists.
     * @param artistname the artists name to check for
     * @param albumname the name of the album
     * @return true if such an album exists for the given artist, false otherwise.
     */
    bool containsAlbum( const QString& artistname, const QString& albumname ) const;

    /**
     * Returns a list with all the tracks in the database.
     * @param buffer the list to put the tracks into.
     * @return the pointer to the given filled list
     */
    itunesdb::TrackPtrList * getAllTracks( itunesdb::TrackPtrList& buffer ) const;

    /**
     * Returns a list with all the tracks in the database
     */
    TrackIterator getAllTracks() const;

    /**
     * Returns an album by the given artistname and title.
     * @param artistname name of the artist of the album in question
     * @param albumname name of the album
     * @param buffer the list the tracks of the album will be put to
     * @return a pointer to the given @c buffer or NULL if no such album could by found
     */
    itunesdb::TrackPtrList *
        getAlbum( const QString& artistname, const QString& albumname, itunesdb::TrackPtrList& buffer ) const;


    /**
     * Returns all the Tracks by the given artist.
     * @param artistname the name of the artist to look for.
     * @param buffer the list to put the found tracks into.
     * @return the pointer to the given buffer or NULL if no such artist exists.
     */
    itunesdb::TrackPtrList *
        getTracksByArtist( const QString& artistname, itunesdb::TrackPtrList& buffer ) const;


    /**
     * Returns all tracks with the given string in its metadata (artist, album, title or composer)
     * @param someString the string to look for
     * @param buffer the list to put the found tracks into
     * @param caseSensitive set to false if uppercase/lowercase should be ignored
     * @return the pointer to the given list or NULL if no such tracks exist.
     */
    itunesdb::TrackPtrList *
        getTracksContaining( const QString& someString, itunesdb::TrackPtrList& buffer, bool caseSensitive = true ) const;


    /**
     * Returns a list with all playlist names
     * @param buffer the list to put the playlisttitles into
     * @return the pointer to the given buffer
     */
    QStringList * getPlaylistTitles( QStringList& buffer ) const;

    /**
     * Returns a playlist by the given playlist title.
     * @param playlisttitle title of the playlist in question
     * @return a pointer to the playlist found or NULL if no such playlist exists
     */
    ITunesDBPlaylist * getPlaylistByTitle(const QString& playlisttitle) const;

    /**
     * Returns the playlist by the given playlistID.
     * @param playlistID the DB ID of the desired playlist
     * @return the playlist found or NULL if no such playlist could be found.
     */
    ITunesDBPlaylist * getPlaylistByID( Q_UINT64 playlistID ) const;

    /**
     * Removes all the tracks of the given artist (an thus the artist itself)
     * from the database.
     * @param artistname the name of the artist
     * @return true if tracks were found and removed, false otherwise.
     */
    bool removeArtist(const QString& artistname);

    /**
     * Renames an album with the given title and artist to newtitle, newartist by
     * changing the artist and album properties of all the tracks found by the
     * given artist/album combination to newartist/newalbum.
     * @param artist the artist to look for
     * @param title the album to look for
     * @param newartist the new artist name to set. Give an empty or QString::null string if the artist property should remain unchanged.
     * @param newtitle the new album title to set. Give an empty or QString::null string if the album property should remain unchanged.
     * @return true if tracks were found and changed, false otherwise
     */
    bool renameAlbum(const QString& artist, const QString& title, const QString& newartist, const QString& newtitle);

    /**
     * adds a new playlist to the database
     */
    bool addPlaylist( ITunesDBPlaylist * playlist );

    /**
     * Removes the playlist with the given title
     * @param title the title of the playlist to be removed
     * @param delete_instance if set to true the element gets deleted during removal
     * @return true if successful, otherwise false
     */
    bool removePlaylist( const QString& title, bool delete_instance);

    /**
     * Updates all Smart Playlists in this itunesdb instance.
     */
    void updateSmartPlaylists();

    /**
     * Removes the Track with the given trackid from the database.
     * This also removes all playlist references to this track
     * @param trackid ID of the track to be deleted
     * @param delete_instance if set to true (default) the Track gets deleted during removal
     * @return the trackid of the Track or 0 if no such Track exists
     */
    Q_UINT32 removeTrack(Q_UINT32 trackid, bool delete_instance = true);

    /**
     * Wipes all data from the container.
     */
    void clear();

    /**
     * returns true if something in the container's control has been changed.
     * a change to a Playlist or a Track for example will not set the change flag yet.
     */
    bool isDirty() const;

    /**
     * Returns the number of tracks
     */
    Q_UINT32 getNumTracks() const;

    /**
     * returns the number of playlists
     */
    Q_UINT32 getNumPlaylists() const;

    /**
     * Lock control functions to prevent concurrent access to the Database from different slaves
     */
    void lock(bool write_lock);
    void unlock();
    bool isLocked() const;

    uint lastModified() { return timestamp.toTime_t(); }

    /**
     * sets the title of the mainlist
     */
    void setTitle( const QString& title );

    /**
     * Returns the title of the mainlist
     */
    const QString& getTitle() const { return m_Title; }

    bool hasPodcasts() const { return hasPodcastsFlag; }

    /**
     * Returns the next free trackID for a new track
     */
    Q_UINT32 getNextFreeTrackID() const;

    /**
     * Returns the next available DBID.
     */
    Q_UINT64 getNextFreeDBID() const;

    /**
     * translates the pathinfo stored on the ipod to a real path info and adds the
     * ipodBase path so we can access the real file. The caller needs to check if
     * the file actually exists.
     * @return the path pointing to the file.
     */
    QString getFileForPathInfo( QString pathinfo ) const;

    /**
     * Returns the ipods base directory.
     */
    const QString& getIPodBase() const {
        return m_ipodSysInfo.getDeviceMountpoint();
    }

protected:

    TrackIDMap m_TrackIdMap;
    TrackList m_TrackList;
    QString m_Title;

    /**
     * The container that holds all the playlist for this iTunesDB instance.
     */
    class PlaylistContainer : public QPtrList<ITunesDBPlaylist> {
    public:
        PlaylistContainer() : QPtrList<ITunesDBPlaylist>() {}
        virtual ~PlaylistContainer() { clear(); }
        virtual int compareItems( QPtrCollection::Item s1, QPtrCollection::Item s2 ) {
            return (*(ITunesDBPlaylist*)s1).getTitle().lower().localeAwareCompare( (*(ITunesDBPlaylist*)s2).getTitle().lower() );
        }
    } * playlists;

    bool m_dirty;
    void setDirty( bool dirtyFlag = true );

    Q_UINT32 maxtrackid;
    Q_UINT64 mMaxDBID;

    ITunesDBTrack * insertTrackToDataBase( ITunesDBTrack * track );

private:

    void removeFromAllPlaylists( Q_UINT32 trackid );
    IPodSysInfo& m_ipodSysInfo;
    QFile itunesdbfile;
    QFile itunessdfile;
    QDateTime timestamp;

    bool hasPodcastsFlag;
    bool locked;
};


#endif
