 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/
#ifndef ITUNESDBPLAYLIST_H
#define ITUNESDBPLAYLIST_H

#include <qlist.h>

#include "utils.h"
#include "listitem.h"
#include "playlistitem.h"
#include "smartplaylistrules.h"

namespace itunesdb {

/**
 * This class represents a playlist in the database.<br>
 * It allows to read, create, manipulate and write back all kinds of playlists. It's part of
 * the itunesdb namespace which is intended to be used for plain reading/writing the database.
 * More enhanced features like sorting or updating smart playlists is implemented in the
 * @c ITunesDB related implementation @c ITunesDBPlaylist <br>
 * For normal playlists the @c addPlaylistItem( Q_UINT32, int ) method can be used to
 * add tracks to the playlist, @c removeAll( Q_UINT32 ) and @c removeTrackAt( uint )
 * will remove them.
 * For editing the rules of smart playlists use the @c getSmartPlaylistRules() method to retrieve
 * the @c SmartPlaylistRuleSet and use its methods for manipulating the rules.
 * To create a new smart playlist create a new playlist and use @c enableSmartPlaylist() method to
 * enable smart playlist behaviour. This method also returns the still empty
 * @c SmartPlaylistRuleSet ready for adding new rules.
 *
 * @see ITunesDBPlaylist
 * @see SmartPlaylistRuleSet
 * @author Michael Schulze
 */
class Playlist : public ListItem
{
    friend class ItunesDBParser;
    friend class ItunesDBWriter;

public:
    enum Sortorder {
        SORTORDER_MANUAL = 1,
        SORTORDER_TITLE = 3,
        SORTORDER_ALBUM = 4,
        SORTORDER_ARTIST = 5,
        SORTORDER_BITRATE = 6,
        SORTORDER_GENRE = 7,
        SORTORDER_FILETYPE = 8,
        SORTORDER_TIME_MODIFIED = 9,
        SORTORDER_TRACK_NR = 10,
        SORTORDER_SIZE = 11,
        SORTORDER_TIME = 12,
        SORTORDER_YEAR = 13,
        SORTORDER_SAMPLERATE = 14,
        SORTORDER_COMMENT = 15,
        SORTORDER_TIME_ADDED = 16,
        SORTORDER_EQUALIZER = 17,
        SORTORDER_COMPOSER = 18,
        SORTORDER_PLAYCOUNT = 20,
        SORTORDER_TIME_PLAYED = 21,
        SORTORDER_CD_NR = 22,
        SORTORDER_RATING = 23,
        SORTORDER_RELEASE_DATE = 24,
        SORTORDER_BPM = 25,
        SORTORDER_GROUPING = 26,
        SORTORDER_CATEGORY = 27,
        SORTORDER_DESCRIPTION = 28
    } ItdbPlaylistSortOrder;

    class TrackList_T : public QList<PlaylistItem> {
        typedef QTPOD_SHARED_PTR_IMPL_DEF<PlaylistItemComparator> ComparatorPtr;
        ComparatorPtr m_comparator;
    public:
        TrackList_T();
        ~TrackList_T();
        void setComparator( PlaylistItemComparator * comparator );
        virtual int compareItems( QPtrCollection::Item s1, QPtrCollection::Item s2 );
    };

    class ConstIterator {
    protected:
        const TrackList_T& m_list;
        TrackList_T::const_iterator m_next;
        TrackList_T::const_iterator m_current;
        friend class itunesdb::Playlist;
        void setPos( TrackList_T::iterator pos ) {
            m_next = pos;
            m_current = pos;
        }
        ConstIterator(const TrackList_T& list)
            : m_list(list), m_next( m_list.begin() ), m_current( m_next ) {}
    public:
        bool hasNext() {
            return m_next != m_list.end();
        }
        bool atEnd() {
            return m_current != m_list.end();
        }
        const PlaylistItem * current() {
            return *m_current;
        }
        const PlaylistItem * next() {
            if ( hasNext() ) {
                m_current = m_next++;
                return *m_current;
            } else {
                return NULL;
            }
        }
    };

    class Iterator {
    protected:
        TrackList_T& m_list;
        TrackList_T::iterator m_next;
        TrackList_T::iterator m_current;
        friend class itunesdb::Playlist;
        void setPos( TrackList_T::iterator pos ) {
            m_next = pos;
            m_current = pos;
        }
        Iterator(TrackList_T& list)
            : m_list(list), m_next( m_list.begin() ), m_current( m_next ) {}
    public:
        bool hasNext() {
            return m_next != m_list.end();
        }
        bool atEnd() {
            return m_current == m_list.end();
        }
        PlaylistItem * current() {
            return *m_current;
        }
        PlaylistItem * next() {
            if ( hasNext() ) {
                m_current = m_next++;
                return *m_current;
            } else {
                return NULL;
            }
        }
    };

    Playlist();
    virtual ~Playlist();

    /**
     * Returns the title of the playlist
     * @return the title of the playlist
     */
    const QString& getTitle() const;

    /**
     * Sets the title of the playlist
     * @param newtitle the new title
     */
    void setTitle( const QString& newtitle);

    /**
     * Returns the sort order field of the playlist
     * @return the sort order field of the playlist
     */
    Sortorder getSortOrder() const;


    /**
     * Returns true if the playlist contains the specified trackid.
     */
    bool contains( Q_UINT32 trackID ) const;

    /**
     * Adds a new PlaylistItem at the specified position to the list and sets
     * the dirty flag to true. The position field only works if the playlists
     * sort order is set to "manual" which is the only sort order implemented
     * by this class, but may be ignored if you use the @c ITunesDBPlaylist
     * implementation and set the sort order to something other than manual.
     * @param trackid the id of the track to add. Doesn't check if the track actually exists
     * @param position the position field as found in the itunesdb mhip playlist item, or -1 if the item should just be appended to the list
     * @return the index of the item in the list or -1 if the item could not be added or the item was NULL
     * @see @c ITunesDBPlaylist for a full implementation of itunesDBs playlists
     */
    int addPlaylistItem( Q_UINT32 trackid, int position = -1 );

    /**
     * Removes the track at the position the given Iterator points to and returns the
     * former trackid at that position. If the Iterator doesn't point to an element in this
     * list PLAYLISTITEM_INVALID (as defined in playlistitem.h) is returned.
     * After the operation the given Iterator will point to the next element.
     */
    Q_UINT32 removeTrackAt( Iterator& pos );

    /**
     * Removes the track at the given position and returns the
     * former trackid at that position. If the given position is out of range
     * PLAYLISTITEM_INVALID (as defined in playlistitem.h) is returned.
     */
    Q_UINT32 removeTrackAt( uint pos );

    /**
     * Removes all occurences of the track with the given ID. Returns true if an element
     * with the given trackid existed and was removed, false otherwise.
     */
    bool removeAll( Q_UINT32 trackid );

    /**
     * Returns the trackID at the given position or PLAYLISTITEM_INVALID
     * (as defined in playlistitem.h) if the given position is either negative
     * or above the last element in the list.
     */
    virtual Q_UINT32 getTrackIDAt( uint pos );

    /**
     * Returns an iterator over all tracks in this playlist. The iterator
     * works like a java iterator - you can check if there are any elements
     * left with the @c Iterator::hasNext() method and get the next element with the
     * @c Iterator::next() method.
     */
    virtual Iterator getTrackIDs();

    /**
     * The const version of the above method.
     * Returns an iterator over all tracks in this playlist.
     */
    virtual ConstIterator getTrackIDs() const;

    /**
     * Returns the number of tracks in this playlist
     * @return the number of tracks in this playlist
     */
    virtual uint getNumTracks() const;

    /**
     * Returns the unique DB id for this playlist
     * @return the unique DB id for this playlist
     */
    Q_UINT64 getID() const { return m_id; }

    /**
     * Sets the unique DB id for this playlist
     * @param id the unique DB id for this playlist
     */
    void setID( Q_UINT64 id );

    /**
     * Sorts the playlistItems by position.
     */
    void sort();

    /**
     * Empties the playlist.
     * After a call to this method the @c getNumTracks() method returns 0.
     */
    void clear();

    /**
     * Returns true if this playlist is hidden.
     */
    virtual bool isHidden() const { return hidden; }

    /**
     * Enables Smart Playlist Behaviour by creating a SmartPlaylistRuleSet.
     */
    SmartPlaylistRuleSet& enableSmartPlaylist();

    /**
     * Returns true if this playlist is a smart playlist.
     */
    bool isSmartPlaylist() const;

    /**
     * Discards smart playlist data thus making it a normal playlist.
     * The items in this playlist will be kept.
     */
    void discardSmartPlaylistData();

    /**
     * Returns the SmartPlaylistRuleSet making up this smart playlist, or NULL
     * IF this playlist isn't a SmartPlaylist. All smart playlist behaviour can
     * be / is controlled by the returned object.
     */
    SmartPlaylistRuleSet * getSmartPlaylistRules() const;

    /**
     * Returns true if this playlist is a podcast list
     */
    bool isPodcastList() const { return mIsPodcast; }

    /**
     * Adds all tracks delivered by the given Iterator.
     */
    template <class JavaLikeTrackIterator>
        void addAll( JavaLikeTrackIterator trackIter ) {
            bool changed = false;
            while ( trackIter.hasNext() ) {
                addPlaylistItem( trackIter.next()->getID() );
                changed = true;
            }
            if ( changed ) {
                setDirty();
            }
        }

    /**
     * Adds all tracks delivered by the given Iterator.
     */
    template <class CPPStyleTrackIterator>
        void addAll( CPPStyleTrackIterator trackIter, CPPStyleTrackIterator end ) {
            bool changed = false;
            for ( ; trackIter != end; trackIter++ ) {
                addPlaylistItem( (*trackIter)->getID() );
                changed = true;
            }
            if ( changed ) {
                setDirty();
            }
        }

protected:

    Playlist( const Playlist& );

    /**
     * Creates a new PlaylistItem. Override this method if you need your own PlaylistItem implementation.
     */
    virtual PlaylistItem * createNewItem( Q_UINT32 trackid );

    virtual QDataStream & readFromStream( QDataStream & instream, bool * ok = NULL );

    virtual QDataStream & writeToStream ( QDataStream & outstream, bool isMainlist );

    /**
     * override from ListItem::doneAddingData()
     */
    virtual void doneAddingData();

    /**
     * Adds a new playlist item to the internal list.
     */
    virtual int addPlaylistItem( PlaylistItem * item );

    virtual SmartPlaylistRuleSet * createNewSplRuleSet();

    void fillSplHeaderBuffer( QByteArray& buffer ) const;

    void fillSplRulesBuffer( QByteArray& buffer ) const;

    virtual void readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen );
    virtual uint writeNonStringMHODs( QDataStream& outstream ) const;

    void writeData( QByteArray& data, bool isMainlist) const;
    void writeTitle( QDataStream& stream ) const;
    void writeLongPlaylist( QDataStream& stream ) const;
    void writeTracks( QDataStream& stream ) const;

    TrackList_T m_tracklist;
    Q_UINT32 hidden;
    Q_UINT32 timeStamp;     // some timestamp
    Q_UINT64 m_id;          // playlist ID
    Q_UINT32 unk3;
    Q_UINT16 mIsPodcast;    // podcast flag
    Sortorder m_order;

    SmartPlaylistRuleSet * mSplRuleSet;

};

}

#endif
