 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef ITUNESDBSMARTPLAYLISTRULES_H
#define ITUNESDBSMARTPLAYLISTRULES_H

#include <qstring.h>
#include <vector>

#define ACTION_NOT_BIT 0x02000000
#define ACTION_STRING_BIT 0x01000000

#define SPL_VALUE_NOW 0x2dae2dae2dae2daeLL

#define IN_THE_LAST_SECONDS 1
#define IN_THE_LAST_DAYS    86400
#define IN_THE_LAST_WEEKS   604800

namespace itunesdb {

/**
    Defines Smart Playlist Rules

    @author Michael Schulze
*/


enum SPLStringField {
    FIELD_NAME = 0x02,
    FIELD_ALBUM = 0x03,
    FIELD_ARTIST = 0x04,
    FIELD_GENRE = 0x08,
    FIELD_KIND = 0x09,
    FIELD_COMMENT = 0x0e,
    FIELD_COMPOSER = 0x12,
    FIELD_GROUPING = 0x27,
    FIELD_DESCRIPTION = 0x36,
    FIELD_CATEGORY = 0x37
};


enum SPLUIntField {
    FIELD_BITRATE = 0x05,
    FIELD_SAMPLERATE = 0x06,
    FIELD_YEAR = 0x07,
    FIELD_DATEMODIFIED = 0x0a,
    FIELD_TRACKNUMBER = 0x0b,
    FIELD_SIZE = 0x0c,
    FIELD_TIME = 0x0d,
    FIELD_DATEADDED = 0x10,
    FIELD_PLAYCOUNT = 0x16,
    FIELD_LASTPLAYED = 0x17,
    FIELD_DISCNUMBER = 0x18,
    FIELD_RATING = 0x19,
    FIELD_COMPILATION = 0x1f,
    FIELD_BPM = 0x23,
    FIELD_PLAYLISTID = 0x28,
    FIELD_PODCAST = 0x29
};


enum SPLStringAction {
    ACTION_IS_STRING = 0x01000001,
    ACTION_CONTAINS = 0x01000002,
    ACTION_STARTSWITH = 0x01000004,
    ACTION_ENDSWITH = 0x01000008,
    ACTION_IS_NOT_STRING = ACTION_IS_STRING ^ ACTION_NOT_BIT,
    ACTION_DOESNT_CONTAIN = ACTION_CONTAINS ^ ACTION_NOT_BIT,
    ACTION_DOESNT_STARTWITH = ACTION_STARTSWITH ^ ACTION_NOT_BIT,
    ACTION_DOESNT_ENDWITH = ACTION_ENDSWITH ^ ACTION_NOT_BIT
};


enum SPLUIntAction {
    ACTION_IS_UINT = 0x00000001,
    ACTION_GREATER_THAN = 0x00000010,
    ACTION_GREATER_OR_EQUAL = 0x00000020,
    ACTION_LESS_THAN = 0x00000040,
    ACTION_LESS_OR_EQUAL = 0x00000080,
    ACTION_IN_RANGE = 0x00000100,
    ACTION_IN_THE_LAST = 0x00000200,
    ACTION_IS_NOT_UINT = ACTION_IS_UINT ^ ACTION_NOT_BIT,
    ACTION_NOT_GREATER = ACTION_GREATER_THAN ^ ACTION_NOT_BIT,
    ACTION_NOT_GREATER_OR_EQUAL = ACTION_GREATER_OR_EQUAL ^ ACTION_NOT_BIT,
    ACTION_NOT_LESS = ACTION_LESS_THAN ^ ACTION_NOT_BIT,
    ACTION_NOT_LESS_OR_EQUAL = ACTION_LESS_OR_EQUAL ^ ACTION_NOT_BIT,
    ACTION_NOT_IN_RANGE = ACTION_IN_RANGE ^ ACTION_NOT_BIT,
    ACTION_NOT_IN_THE_LAST = ACTION_IN_THE_LAST ^ ACTION_NOT_BIT
};


enum SPLLimitSort {
    LIMIT_SORT_RANDOM        = 0x02,
    LIMIT_SORT_SONG_NAME     = 0x03,
    LIMIT_SORT_ALBUM         = 0x04,
    LIMIT_SORT_ARTIST        = 0x05,
    LIMIT_SORT_GENRE         = 0x07,
    LIMIT_SORT_RECENTLYADDED = 0x10,
    LIMIT_SORT_PLAYCOUNT     = 0x14,
    LIMIT_SORT_LASTPLAYED    = 0x15,
    LIMIT_SORT_RATING        = 0x17
};


enum SPLLimitType {
    LIMIT_TYPE_MINUTES   = 1,
    LIMIT_TYPE_MEGABYTES = 2,
    LIMIT_TYPE_SONGS     = 3,
    LIMIT_TYPE_HOURS     = 4,
    LIMIT_TYPE_GIGABYTES = 5
};


/**
 * This class represents a single smart playlist rule.
 */
class SmartPlaylistRule {

    friend class SmartPlaylistRuleSet;

    Q_UINT32 mField;
    Q_UINT32 mAction;

    const QString mString;

    const Q_UINT64 mFromValue;
    const Q_UINT64 mToValue;

    const Q_INT64 mFromDate;
    const Q_INT64 mToDate;

    const Q_UINT64 mFromUnits;
    const Q_UINT64 mToUnits;

    Q_UINT64 mCalculatedFrom;
    Q_UINT64 mCalculatedTo;

public:

    virtual ~SmartPlaylistRule();

    /**
     * Returns the action identifier for this rule
     */
    Q_UINT32 getAction() const {
        return mAction;
    }

    /**
     * Returns the field identifier for this rule
     */
    Q_UINT32 getField() const {
        return mField;
    }

    /**
     * Returns the String data for String based rules.
     */
    const QString& getString() const {
        return mString;
    }

    /**
     * Returns the "from" value as set in the itunesdb rule definition.
     * If you need the "from" value actually compared to the desired fields
     * use the @c calculatedFrom() method.
     */
    Q_UINT64 getFromValue() const {
        return mFromValue;
    }

    /**
     * Returns the "to" value as set in the itunesdb rule definition.
     * If you need the "to" value actually compared to the desired fields
     * use the @c calculatedTo() method.
     */
    Q_UINT64 getToValue() const {
        return mToValue;
    }

    /**
     * Returns the "from date" as set in the itunesdb rule definition.
     * If you need the "from" value actually compared to the desired fields
     * use the @c calculatedFrom() method.
     */
    Q_INT64 getFromDate() const {
        return mFromDate;
    }

    /**
     * Returns the "to date" as set in the itunesdb rule definition.
     * If you need the "to" value actually compared to the desired fields
     * use the @c calculatedTo() method.
     */
    Q_INT64 getToDate() const {
        return mToDate;
    }

    /**
     * Returns the "from units" as set in the itunesdb rule definition and
     * defines the units the fromDate is measured in.
     */
    Q_UINT64 getFromUnits() const {
        return mFromUnits;
    }

    /**
     * Returns the "to units" as set in the itunesdb rule definition and
     * defines the units the toDate is measured in.
     */
    Q_UINT64 getToUnits() const {
        return mToUnits;
    }

    /**
     * Returns the "from" value actually compared to the fields in question.
     * The value is calculated via ( fromValue + fromDate * fromUnits ).
     */
    Q_UINT64 calculatedFrom() const;

    /**
     * Returns the "to" value actually compared to the fields in question.
     * The value is calculated via ( toValue + toDate * toUnits ).
     */
    Q_UINT64 calculatedTo() const;

    /**
     * Writes this SPL TrackPredicate as a SPL Rule to the given stream
     * and returns the number of bytes written
     */
    uint writeToStream( QDataStream& outStream ) const;

    /**
     * Returns true if the rule is a string based rule
     */
    bool isStringBasedRule() const {
        return getAction() & ACTION_STRING_BIT;
    }

    /**
     * Returns a textual represenation of this rule (human readable)
     */
    virtual const QString toString() const;

protected:

    /**
     * Constructor to create a string based SmartplaylistRule
     */
    SmartPlaylistRule( SPLStringField field, SPLStringAction action, const QString& testString );

    /**
     * Constructor to create an uint based SmartplaylistRule
     */
    SmartPlaylistRule( SPLUIntField field, SPLUIntAction action,
                       Q_UINT64 fromValue, Q_UINT64 toValue,
                       Q_INT64 fromDate, Q_INT64 toDate,
                       Q_UINT64 fromUnits, Q_UINT64 toUnits );

    static QString getActionString( uint action );

    static QString getFieldString( uint field );

    static Q_UINT64 calculateTestValue( Q_UINT64 value, Q_INT64 date, Q_UINT64 units );

};


typedef std::vector<SmartPlaylistRule*> RuleList;

/**
 * This class represents all the rules, limits and switches making up a Smart Playlist.
 * Here is an example on how to create a "recently played" Smart Playlist:
 * <pre><code>
 *  // enable smart playlist behaviour and get the ruleset
 *  SmartPlaylistRuleSet& ruleSet = a_playlist_ptr->enableSmartPlaylist();
 *
 *  // add the rule "last played in the last 2 weeks"
 *  ruleSet.addInTheLastRule( FIELD_LASTPLAYED, false, -2, IN_THE_LAST_WEEKS );
 *
 *  // limit to the first 30 songs, sorted by last played date
 *  ruleSet.setLimits( LIMIT_SORT_LASTPLAYED, LIMIT_TYPE_SONGS, 30 );
 *
 *  // match ANY rule (doesn't really matter since we only have one rule)
 *  ruleSet.setMatchAnyFlag( true );
 *
 *  // enable live update (updates the playlist on the go)
 *  ruleSet.setLiveUpdateFlag( true );
 * </code></pre>
 */
class SmartPlaylistRuleSet {

    friend class Playlist;

public:

    SmartPlaylistRuleSet();
    virtual ~SmartPlaylistRuleSet();

    /**
     * Adds a string based rule to the rule set and returns true if successful.
     * @param field field ID to identify the track field in question
     * @param action the compare operation to use when comparing the @c field with @c testString
     * @param testString the String to compare the @c field with
     * @param position the position the rule will be inserted to the rule set. Set to -1 to append the rule to the end
     */
    virtual bool addStringRule( SPLStringField field,
                        SPLStringAction action, const QString& testString, int position = -1 );

    /**
     * Adds a uint based rule to the rule set and returns true if successful.<br>
     * For rules with plain integer type fields like BPM, rating or play count the
     * from and to values need to be set to the range or value you care about.
     * See @c addUIntRule( SPLUIntField, SPLUIntAction, Q_UINT64, Q_UINT64) for
     * a simpler way to achieve this.<br>
     * For timestamp type rules like "last played in the last 2 weeks" you
     * need to set the from and to values to the offset of the time range to
     * care about, or to SPL_VALUE_NOW if the desired date offset is "now".
     * The from and to values declare the time range and the units the time
     * is measured in.<br>
     * Now, if you need to declare the aforementioned rule you set the time
     * range [from-to] to ["now - 2 weeks" - "now"] by setting both the from
     * and to values to SPL_VALUE_NOW, set the fromDate to -2 the to date to 0
     * and the units to IN_THE_LAST_WEEKS as fromUnits and 1 as toUnits.
     * @see @c addInTheLastRule() for a better way to create these kinda rules.
     * @see http://ipodlinux.org/ITunesDB#Smart_Playlist_Rules for more details
     * @param field field ID to identify the track field in question
     * @param action the compare operation to use
     * @param fromValue the "from" value
     * @param toValue the "to" value
     * @param fromDate the "from" date
     * @param toDate the "to" date
     * @param fromUnits the units the from value is measured in (@c fromValue gets multiplied with the units)
     * @param toUnits the units the to value is measured in (@c toValue gets multiplied with the units)
     * @param position the position the rule will be inserted to the rule set. Set to -1 to append the rule to the end
     */
    virtual bool addUIntRule( SPLUIntField field, SPLUIntAction action,
                              Q_UINT64 fromValue, Q_UINT64 toValue,
                               Q_INT64 fromDate,   Q_INT64 toDate,
                              Q_UINT64 fromUnits, Q_UINT64 toUnits, int position = -1 );

    /**
     * Convenience method. This creates a UInt based rule setting the from and
     * to values as provided and using sane defaults for the rest.
     * @param field field ID to identify the track field in question
     * @param action the compare operation to use
     * @param fromValue the "from" value
     * @param toValue the "to" value
     * @param position the position the rule will be inserted to the rule set. Set to -1 to append the rule to the end
     */
    bool addUIntRule( SPLUIntField field, SPLUIntAction action,
                              Q_UINT64 fromValue, Q_UINT64 toValue, int position = -1 ) {
        return addUIntRule( field, action, fromValue, toValue, 0, 0, 1, 1, position );
    }


    /**
     * Convenience method. This creates a UInt based "In The Last" rule with the given parameters
     * @param field field ID, to identify the field in question.
     * @param negateRule if true, this negates the rule to "Not In The Last"
     * @param value the time to evaluate; like in the last [WEEKS|DAYS|HOURS|etc]
     * @param units the units the value is mesaured in; base is seconds. The Predefined values IN_THE_LAST_[SECONDS|DAYS|WEEKS] may be used
     * @param position the position the rule will be inserted to the rule set. Set to -1 to append the rule to the end
     */
    bool addInTheLastRule( SPLUIntField field, bool negateRule, Q_INT64 value, Q_UINT64 units, int position = -1 ) {
        return addUIntRule( field,
                    negateRule ? ACTION_NOT_IN_THE_LAST : ACTION_IN_THE_LAST,
                    SPL_VALUE_NOW, SPL_VALUE_NOW, value, 0, units, 1, position );
    }


    /**
     * Returns the Rule at the specified position or NULL if no such rule exists.
     * @param position the position
     * @return the rule at the position or NULL.
     */
    const SmartPlaylistRule * getRuleAt( uint position ) {
        return mRules.at( position );
    }

    /**
     * Removes the rule at the specified position. The list will be compacted,
     * so all the rules after the rule to be removed will be moved one position
     * towards the start of the list.
     * @param position the position to remove the rule from
     * @return true if the operation was successfull, false otherwise.
     */
    bool removeRuleAt( uint position );

    /**
     * Removes all the rule from this RuleSet
     */
    void clear();


    /**
     * Returns true if the playlist may be updated by the iPod itself.
     */
    bool isLiveUpdate() const {
        return mLiveUpdate;
    }


    /**
     * Sets the live update flag.
     */
    void setLiveUpdateFlag( bool liveupdate );

    /**
     * Returns true if the smart playlist checks the rules
     */
    bool isCheckRules() const {
        return mCheckRules;
    }

    /**
     * Set the checkRules flag.
     */
    void setCheckRulesFlag( bool checkRules );

    /**
     * Returns true if the limits will/need_to be checked
     */
    bool isCheckLimits() const {
        return mCheckLimits;
    }

    /**
     * Sets the limit for this smart playlist and enable limit checking
     */
    virtual void setLimits( SPLLimitSort limitSort, SPLLimitType limitType, Q_UINT32 limitValue );

    /**
     * enables or disables limit checking.
     */
    void enableLimits( bool enable );

    /**
     * returns the limit sort field
     */
    SPLLimitSort getLimitSort() const {
        return mLimitSort;
    }

    /**
     * Returns the limit type
     */
    SPLLimitType getLimitType() const {
        return mLimitType;
    }

    /**
     * Returns true if only one rule needs to match.
     */
    bool isMatchAny() const {
        return mMatchAny;
    }

    /**
     * Sets the "match any" flag
     */
    void setMatchAnyFlag( bool matchAny );

    /**
     * Returns true if only "checked" tracks should be matched against the rules.
     */
    bool checkedTracksOnly() const {
        return mCheckedOnly;
    }

    /**
     * sets the checkedTracksOnlyFlag.
     * @param checkedOnly If set to true only "checked" tracks should be matched against the rules.
     */
    void setCheckedTracksOnlyFlag( bool checkedOnly );

    /**
     * Returns the rules making up this smart playlist rule set.
     */
    const RuleList& getRules() const;

    /**
     * Returns the number of rules
     */
    uint getNumRules() const;

    /**
     * Returns a String representation of this rule set.
     */
    QString toString() const;

protected:

    // do not copy the rules
    SmartPlaylistRuleSet( const SmartPlaylistRuleSet & src );

    /**
     * read Rules from type 51 MHOD data, starting at SLst and returns the number of rules read
     * @param instream the stream to read from positioned at the MHOD data
     * @param length the MHOD data length in bytes (blocklen - headerlen)
     * @return the number of rules read from the stream
     */
    uint readRules( QDataStream & instream, uint length );

    /**
     * reads the SPL Header from a type 50 MHOD and returns if the operation succeeded.
     * @param instream the stream to read from positioned at the MHOD data
     * @param length the MHOD data length in bytes (blocklen - headerlen)
     * @return true if the header got read successfully, false otherwise
     */
    bool readHeader( QDataStream & instream, uint length );

    /**
     * Writes the rules back to the stream and returns the number of bytes written.
     * @param outstream the stream to write the MHOD51 data to.
     * @return the number of bytes written
     */
    uint writeRules( QDataStream & outstream ) const;

    /**
     * Writes the header to the stream as a type 50 MHOD and returns the number of bytes written.
     * @param outstream the stream to write the MHOD50 data to.
     * @return the number of bytes written
     */
    uint writeHeader( QDataStream & outstream ) const;

    /***
     *
     * Header
     *
     ***/

    bool mMatchAny;

    Q_UINT32 unk5;

    Q_UINT8 mLiveUpdate;
    Q_UINT8 mCheckLimits;
    Q_UINT8 mCheckRules;
    Q_UINT8 mCheckedOnly;
    Q_UINT8 mReverseLimitSort;
    SPLLimitSort mLimitSort;
    SPLLimitType mLimitType;
    Q_UINT32 mLimitValue;

    // Attention: copy constructor exists

    RuleList mRules;

    virtual SmartPlaylistRule * createStringRule( SPLStringField field,
                                          SPLStringAction action, const QString& testString );

    virtual SmartPlaylistRule * createUIntRule(  SPLUIntField field, SPLUIntAction action,
                                         Q_UINT64 fromValue, Q_UINT64 toValue,
                                         Q_INT64  fromDate,  Q_INT64 toDate,
                                         Q_UINT64 fromUnits, Q_UINT64 toUnits );

private:

    SmartPlaylistRule * readSingleRule( QDataStream & instream, uint * bytesRead = NULL );

};


}

#endif
