 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/
#ifndef ITUNESDBTRACK_H
#define ITUNESDBTRACK_H

#include <qcstring.h>
#include <qdatetime.h>
#include <qlist.h>
#include <qvaluevector.h>

#include "utils.h"
#include "listitem.h"

namespace itunesdb {


/**
represents a track

@author Michael Schulze
*/
class Track : public ListItem {

    friend class ItunesDBParser;
    friend class ItunesDBWriter;

public:

    Track( Q_UINT32 trackID = 0 );

    virtual ~Track();

    /**
     * Returns the key for the track in the ITunesDB database
     * @return the key for the track in the ITunesDB database
     */
    Q_UINT32 getID() const;

    /**
     * Returns the album the track is released for
     * @return the albumname
     */
    const QString & getAlbum() const;

    /**
     * Returns the artist
     * @return the artists name
     */
    const QString & getArtist() const;

    QString getPlainArtistName() const;

    /**
     * Returns the category
     * @return the category
     */
    const QString& getCategory() const;

    /**
     * Returns the comment set for the track
     * @return the comment
     */
    const QString & getComment() const;

    /**
     * Returns the title of the track
     * @return the title
     */
    const QString & getTitle() const;

    /**
     * Returns the genre of the track
     * @return the genre
     */
    const QString & getGenre() const;

    /**
     * Returns the pathname of the track as stored in the itunesdb.
     * Be aware that the path is relative to the ipod base directory
     * and the path separator is ":".
     * @return pathname of the track
     */
    const QString & getPathInfo() const;

    /**
     * Returns the composer
     * @return the composer
     */
    const QString & getComposer() const;

    /**
     * Retunrs the EQ Setting
     */
    const QString & getEqualizerSetting() const;

    /**
     * Returns some descriptional text for this piece of audio data.
     */
    const QString& getDescription() const;

    /**
     * Returns the grouping info for this song.
     */
    const QString& getGrouping() const;

    const QString& getFileFormatDesc() const;

    /**
     * Returns the DBID which refers to additional media associated to this track (like cover art)
     * @return the DBID
     */
    Q_UINT64 getDBID() const;

    /**
     * Returns the rating for this track as read from the itunesdb.
     * @return the rating for this track as read from the itunesdb.
     */
    unsigned char getRating() const;

    /**
     * Returns the last modified date for this track.
     * @return the last modified date for this track.
     */
    Q_UINT32 getLastModified() const;

    /**
     * Returns the size of the file associated with this track
     * @return the filesize
     */
    Q_UINT32 getFileSize() const;

    /**
     * Returns the length of the track in milliseconds
     * @return the length of the track in milliseconds
     */
    Q_UINT32 getTrackLength() const;

    /**
     * Returns the number of the track
     * @return the number of the track
     */
    Q_UINT32 getTrackNumber() const;

    /**
     * Returns the number of tracks in the album
     * @return the number of tracks in the album
     */
    Q_UINT32 getTrackCount() const;

    /**
     * Returns the year the song was created
     * @return the year the song was created
     */
    Q_UINT32 getYear() const;

    /**
     * Returns the bitrate
     * @return the bitrate
     */
    Q_UINT32 getBitrate() const;

    /**
     * Returns the sample rate
     * @return the sample rate
     */
    Q_UINT32 getSamplerate() const;

    /**
     * Returns the volume adjust for the track
     * @return the volume adjust
     */
    Q_UINT32 getVolumeAdjust() const;

    /**
     * Returns the number of times the track was played so far
     * @return the play count
     */
    Q_UINT32 getPlayCount() const;

    /**
     * Returns the date the track was last played.
     */
    Q_UINT32 getLastPlayed() const;

    /**
     * Returns the cd number the track is on for albums with more than one CD.
     * @return the cd number the track is on for albums with more than one CD.
     */
    Q_UINT32 getCdNumber() const;

    /**
     * Returns the number of CDs for the album of the track.
     * @return the number of CDs for the album of the track.
     */
    Q_UINT32 getCdCount() const;

    /**
     * Returns the date the track was added to the database/device.
     * @return the date the track was added to the database/device.
     */
    Q_UINT32 getDateAdded() const { return date_added; }

    Q_UINT32 getStartTime() const { return mStartTime; }
    Q_UINT32 getStopTime() const { return mStopTime; }
    Q_UINT32 getSoundCheck() const { return mSoundCheck; }

    unsigned char getVbrFlag() const;

    Q_UINT32 getFileFormatCode() const { return file_format_code; }
    Q_UINT32 getReleaseDate() const { return date_released; }
    unsigned char getMediaType() const { return type; }
    unsigned char getCompilationFlag() const { return compilation; }

    unsigned char isChecked() const { return checked; }

    Q_UINT16 getBPM() const;

    virtual void setDBID( Q_UINT64 id );

    /**
     * Sets the album for this track.
     * @param album the album to set.
     */
    virtual void setAlbum( const QString& album );

    /**
     * Sets the artist for this track.
     * @param artist the artist name to set.
     */
    virtual void setArtist( const QString& artist );

    /**
     * Sets the category for this track.
     * @param category the category to set.
     */
    virtual void setCategory( const QString& category );

    /**
     * Sets the path information for this track. The path is relative to the ipod's base directory.
     * The path separator is ":"
     * @param encodedpath the path information to set.
     */
    virtual void setPathInfo( const QString& encodedpath );

    /**
     * Sets the comment for this track.
     * @param comment the comment to set.
     */
    virtual void setComment( const QString& comment );

    virtual void setCompilationFlag( unsigned char isCompilation );

    /**
     * Sets the description for this track.
     * @param description_txt the description to set.
     */
    virtual void setDescription( const QString& description_txt );

    /**
     * Sets the grouping for this track.
     * @param grouping the grouping to set.
     */
    virtual void setGrouping( const QString& grouping );

    /**
     * Sets the genre for this track.
     * @param genre the genre to set.
     */
    virtual void setGenre( const QString& genre );

    /**
     * Sets the title for this track.
     * @param title the title to set.
     */
    virtual void setTitle( const QString& title );

    /**
     * Sets the file format description for this track.
     * @param fdesc the file format description to set.
     */
    virtual void setFileFormatDesc( const QString& fdesc );

    /**
     * Sets the composer for this track.
     * @param composer the composer to set.
     */
    virtual void setComposer( const QString& composer );

    /**
     * Sets the equalizer setting for this track.
     * @param eqSetting the equalizer preset to set. See the itunesb wiki for details
     */
    virtual void setEqualizerSetting( const QString& eqSetting );

    /**
     * Sets the track number for this track.
     * @param tracknumber the track number to set.
     */
    virtual void setTrackNumber( Q_UINT32 tracknumber );

    /**
     * Sets the the number of tracks in the album for this track.
     * @param numtracks the number of tracks to set.
     */
    virtual void setTrackCount( Q_UINT32 numtracks );

    /**
     * Sets the volume adjust for this track.
     * @param newVolumeAdjust the volume adjust to set.
     */
    virtual void setVolumeAdjust( Q_UINT32 newVolumeAdjust );

    /**
     * Sets the year of release for this track.
     * @param newYear the year to set.
     */
    virtual void setYear( Q_UINT32 newYear );

    /**
     * Sets the start time for this track.
     * @param startTime the start time to set.
     */
    virtual void setStartTime( Q_UINT32 startTime );

    /**
     * Sets the stop time for this track.
     * @param stopTime the stop time to set.
     */
    virtual void setStopTime( Q_UINT32 stopTime );

    /**
     * Sets the soundcheck for this track.
     * @param soundCheck the soundcheck to set.
     */
    virtual void setSoundCheck( Q_UINT32 soundCheck );

    /**
     * Sets the rating for this track.
     * @param rating the rating to set.
     */
    void setRating(unsigned char rating);

    /**
     * Sets the date the track was last played.
     * @param lastplayed the last played date to set.
     */
    void setLastPlayed(Q_UINT32 lastplayed);

    /**
     * Sets the playcount for this track.
     * @param newPlaycount the playcount to set.
     */
    void setPlayCount(Q_UINT32 newPlaycount);

    /**
     * Sets the beats per minute info for this track.
     * @param bpm the beats per minute information to set.
     */
    void setBPM( Q_UINT32 bpm ) { beatsPerMinute = bpm; }

    /**
     * Returns the file extension for this track.
     * @return the file extension for this track.
     */
    QString getFileExtension() const;

    /**
     * Sets the file extension for this track.
     * @param fileextension the file extension to set.
     */
    virtual void setFileExtension( const QString& fileextension );

    /**
     * Copies the metadata from the given source to this Track.
     * The Track class itself is a possible TrackMetaSource, so you
     * can copy metadata from a track to another.
     */
    template <class TrackMetaSource>
    void readFrom( TrackMetaSource& provider ) {
        // Album specific stuff
        setAlbum( provider.getAlbum() );
        setArtist( provider.getArtist() );
        setBPM( provider.getBPM() );
        setCategory( provider.getCategory() );
        setComment( provider.getComment() );
        setCompilationFlag( provider.getCompilationFlag() );
        setComposer( provider.getComposer() );
        setDescription( provider.getDescription() );
        setGrouping( provider.getGrouping() );
        setEqualizerSetting( provider.getEqualizerSetting() );
        setFileFormatDesc(  provider.getFileFormatDesc() );
        setGenre( provider.getGenre() );
        setTitle( provider.getTitle() );
        setTrackCount( provider.getTrackCount() );
        setTrackNumber( provider.getTrackNumber() );
        setVolumeAdjust( provider.getVolumeAdjust() );
        setYear( provider.getYear() );

        // Track related technical stuff
        setLastPlayed( provider.getLastPlayed() );
        setPlayCount( provider.getPlayCount() );
        setRating( provider.getRating() );
        numcds = provider.getCdCount();
        cdnum = provider.getCdNumber();
        file_format_code = provider.getFileFormatCode();
        type = provider.getMediaType();
        tracklen = provider.getTrackLength();
        vbr = provider.getVbrFlag();
        bitrate = provider.getBitrate();
        samplerate = provider.getSamplerate();
        setStartTime( provider.getStartTime() );
        setStopTime( provider.getStopTime() );
        setSoundCheck( provider.getSoundCheck() );

        // Resource related technical stuff
        file_size = provider.getFileSize();
        lastmodified = provider.getLastModified();

        doneAddingData();
    }

    void setRecentlyPlayed( bool playedRecently = true ) { m_recentlyPlayed = playedRecently; }
    bool recentlyPlayed() const { return m_recentlyPlayed; }

protected:

    virtual QDataStream & writeToStream( QDataStream & outstream );
    virtual QDataStream & readFromStream( QDataStream & instream, bool * ok = NULL );

    void writeData( QByteArray& data ) const;

    /**
     * override from ListItem::doneAddingData()
     */
    virtual void doneAddingData();

    Q_UINT32 id;

    // attributes
    Q_UINT32 lastmodified;    // last modified date
    Q_UINT32 file_size;
    Q_UINT32 tracklen;
    Q_UINT32 tracknum;
    Q_UINT32 numtracks;
    Q_UINT32 year;
    Q_UINT32 bitrate;
    Q_UINT32 samplerate;
    Q_UINT32 volumeadjust;
    Q_UINT32 playcount;
    Q_UINT32 last_played_at;
    Q_UINT32 cdnum;
    Q_UINT32 numcds;
    Q_UINT32 file_format_code;
    Q_UINT32 date_added;
    Q_UINT32 date_released;

    Q_UINT32 mStartTime;
    Q_UINT32 mStopTime;
    Q_UINT32 mSoundCheck;

    unsigned char rating;       // rating 0 lowest 100 highest
    unsigned char vbr;          // vbr=1, cbr=0
    unsigned char type;         // mp3=1, aac+audible=0
    unsigned char compilation;  // iscompiltaion=1, 0 otherwise

    Q_UINT64 dbid;
    Q_UINT8  checked;
    Q_UINT8  app_rating;
    Q_UINT16 beatsPerMinute;
    Q_UINT16 mArtworkCount;
    Q_UINT16 unk9;

    // TODO implement podcast flag (together with the whole rest)

private:

    bool m_recentlyPlayed;
    uint m_plainArtistNameStartIdx;
};

/**
 * This interface defines a comparator comparing 2 Tracks and returning a negative int, zero, or
 * a positive int if the first track is smaller, equal or greater than the second.
 *
 * Example: you need to sort a TrackPtrList by Year
 *
 * First create an implementor for TrackComparator like this:
 * <pre><code>
 * class TracksByYearComparator : public TrackComparator {
 * public:
 *     virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) {
 *         return track1.getYear() - track2.getYear();
 *     }
 * };
 * </code></pre>
 * then you call the @c TrackPtrList::setComparator() method of your @c TrackPtrList instance and
 * sort() it.
 * @see itunesdb::TrackComparators for a list of predefined comparators and further examples
 */
class TrackComparator {
public:
    virtual ~TrackComparator() {}
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) = 0;
};

typedef QTPOD_SHARED_PTR_IMPL_DEF<TrackComparator> TrackComparatorPtr;
typedef QTPOD_SHARED_PTR_IMPL_DEF<const TrackComparator> ConstTrackComparatorPtr;

/**
 * This class provides a set of predefined TrackComparator implementations
 */
class TrackComparators {
    class KeepOrder;
    class ByTitle;
    class ByArtist;
    class ByAlbum;
    class ByGenre;
    class ByCategory;
    class ByComment;
    class ByComposer;
    class ByEqualizerSetting;
    class ByDescriptionTxt;
    class ByGrouping;
    class ByBitrate;
    class ByFiletype;
    class ByTimeModified;
    class ByTrackNumber;
    class BySize;
    class ByTrackLength;
    class ByYear;
    class BySampleRate;
    class ByDateAdded;
    class ByPlaycount;
    class ByLastPlayed;
    class ByRating;
    class ByCdNumber;
    class ByReleaseDate;
    class ByBPM;

public:

    /**
     * Comparator to keep the order when calling sort()
     */
    static TrackComparatorPtr KEEP_ORDER;

    /**
     * Comparator to sort a list of tracks by title
     */
    static TrackComparatorPtr BY_TITLE;

    /**
     * Comparator to sort a list of tracks by the artist property
     */
    static TrackComparatorPtr BY_ARTISTONLY;

    /**
     * Comparator to sort a list of tracks by the album property
     */
    static TrackComparatorPtr BY_ALBUMONLY;

    /**
     * Comparator to sort a list of tracks by the genre property
     */
    static TrackComparatorPtr BY_GENREONLY;

    /**
     * Comparator to sort a list of tracks by category
     */
    static TrackComparatorPtr BY_CATEGORY;

    /**
     * Comparator to sort a list of tracks by the comment property
     */
    static TrackComparatorPtr BY_COMMENT;

    /**
     * Comparator to sort a list of tracks by the composer property
     */
    static TrackComparatorPtr BY_COMPOSERONLY;

    /**
     * Comparator to sort a list of tracks by the eq setting property
     */
    static TrackComparatorPtr BY_EQSETTING;

    /**
     * Comparator to sort a list of tracks by the description property
     */
    static TrackComparatorPtr BY_DESCRIPTIONTXT;

    /**
     * Comparator to sort a list of tracks by the grouping property
     */
    static TrackComparatorPtr BY_GROUPING;

    /**
     * Comparator to sort a list of tracks by bitrate
     */
    static TrackComparatorPtr BY_BITRATE;

    /**
     * Comparator to sort a list of tracks by filetype
     */
    static TrackComparatorPtr BY_FILETYPE;

    /**
     * Comparator to sort a list of tracks by modification time
     */
    static TrackComparatorPtr BY_TIMEMODIFIED;

    /**
     * Comparator to sort a list of tracks by track number
     */
    static TrackComparatorPtr BY_TRACKNUMBER;

    /**
     * Comparator to sort a list of tracks by size
     */
    static TrackComparatorPtr BY_SIZE;

    /**
     * Comparator to sort a list of tracks by the tracks length (timewise)
     */
    static TrackComparatorPtr BY_TRACKLENGTH;

    /**
     * Comparator to sort a list of tracks by the year property
     */
    static TrackComparatorPtr BY_YEAR;

    /**
     * Comparator to sort a list of tracks by the sample rate
     */
    static TrackComparatorPtr BY_SAMPLERATE;

    /**
     * Comparator to sort a list of tracks by the date the tracks were added to the database
     * (oldest first).
     */
    static TrackComparatorPtr BY_DATEADDED;

    /**
     * Comparator to sort a list of tracks by the date the tracks were added to the database
     * (most recently first).
     */
    static TrackComparatorPtr BY_RECENTLYADDED;

    /**
     * Comparator to sort a list of tracks by the play count property
     */
    static TrackComparatorPtr BY_PLAYCOUNTONLY;

    /**
     * Comparator to sort a list of tracks by the last played property (most recently last).
     */
    static TrackComparatorPtr BY_LASTPLAYED;

    /**
     * Comparator to sort a list of tracks by the last played property (most recently first).
     */
    static TrackComparatorPtr BY_LASTPLAYED_REVERSE;

    /**
     * Comparator to sort a list of tracks by the rating property.
     */
    static TrackComparatorPtr BY_RATING;

    /**
     * Comparator to sort a list of tracks by the cd number property.
     */
    static TrackComparatorPtr BY_CDNUMBER;

    /**
     * Comparator to sort a list of tracks by release date.
     */
    static TrackComparatorPtr BY_RELEASEDATE;

    /**
     * Comparator to sort a list of tracks by beats per minute.
     */
    static TrackComparatorPtr BY_BPM;

    /**
     * Comparator to sort a list of tracks by album, tracknumber and title
     */
    static TrackComparatorPtr BY_ALBUM;

    /**
     * Comparator to sort a list of tracks by artist, album, tracknumber and title
     */
    static TrackComparatorPtr BY_ARTIST;

    /**
     * Comparator to sort a list of tracks by genre and all the properties BY_ARTIST uses.
     */
    static TrackComparatorPtr BY_GENRE;

    /**
     * Comparator to sort a list of tracks by composer and title
     */
    static TrackComparatorPtr BY_COMPOSER;

    /**
     * Comparator to sort a list of tracks by playcount and all the properties BY_ARTIST uses.
     */
    static TrackComparatorPtr BY_PLAYCOUNT;


    /**
     * Reverses the sort order of the given comparator.
     */
    class ReverseComparator : public TrackComparator {
        TrackComparatorPtr m_delegate;
    public:
        /**
         * Constructs a comparator reversing the sort order of the given comparator.
         */
        ReverseComparator( const TrackComparatorPtr& comparator ) : m_delegate( comparator ) {}
        virtual ~ReverseComparator() {}
        virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) {
            return m_delegate->compare( track2, track1 );
        }
    };


    /**
     * Composition of many comparators. The first Comparator added is the one with the highest
     * priority.
     */
    class ComposingComparator : public TrackComparator {
        typedef QValueVector<TrackComparatorPtr> ComparatorPtrVector;
        ComparatorPtrVector m_comparators;
    public:
        ComposingComparator() {}

        /**
         * Constructs a composing comparator with the given comparator added to the top.
         * The comparator given is owned by this instance so the lifecycle is managed by it.
         */
        ComposingComparator( TrackComparator * first ) {
            if ( first ) {
                m_comparators.append( TrackComparatorPtr( first ) );
            }
        }

        /**
         * Constructs a composing comparator with the given comparator added to the top.
         */
        ComposingComparator( TrackComparatorPtr first ) {
            if ( first ) {
                m_comparators.append( first );
            }
        }
        virtual ~ComposingComparator() {}

        /**
         * Appends the given comparator.
         */
        ComposingComparator * add( TrackComparatorPtr next ) {
            if ( next ) {
                m_comparators.append( next );
            }
            return this;
        }

        /**
         * Appends the given comparator.
         * The comparator given is owned by this instance so the lifecycle is managed by it.
         */
        ComposingComparator * add( TrackComparator * next ) {
            if ( next ) {
                m_comparators.append( TrackComparatorPtr( next ) );
            }
            return this;
        }
        virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 );
    };

    /**
     * Creates a shared ptr to the given comparator.
     */
    static TrackComparatorPtr createPtr( TrackComparator * comparator );
};  // TrackComparators


/**
 * This interface will be used by ITunesDB::findFirstTrackBy and ITunesDB::getTracksBy.
 * Callers to these functions provide a reference to an implementation of this interface
 * and thus provide code to decide if a given track is a valid result.
 *
 * Example: you need to get all the tracks performed at a given year (e.g. 2004) from the database
 *
 * First create an implementor for TrackPredicate like this:
 * <pre><code>
 * class TracksByYearPredicate : public TrackPredicate {
 *     Q_UINT32 m_givenYear;
 * public:
 *     TracksByYearPredicate( Q_UINT32 givenYear ) : m_givenYear( givenYear ) {}
 *     ~TracksByYearPredicate() {}
 *     bool operator() ( const itunesdb::Track& track ) {
 *         return track.getYear() == m_givenYear;
 *     }
 * };
 * </code></pre>
 * then you call the getTracksBy() method of your ITunesDB instance (here itunesdb) with a reference to your predicate
 * <pre><code>
 * TracksByYearPredicate yearPredicate( 2004 );
 * TrackPtrList result;
 * itunesdb.getTracksBy( yearPredicate, result );
 * </code></pre>
 * Now you have all tracks performed at 2004 in your result list.
 * @see ITunesDB::findFirstTrackBy()
 * @see ITunesDB::getTracksBy()
 * }
 */
class TrackPredicate {
public:
    virtual ~TrackPredicate() {}

    /**
     * Implement this method accordingly to the documentation above and (!) the documentation
     * of the function this predicate will be given to.
     */
    virtual bool operator () ( const Track& ) = 0;
};


typedef QTPOD_SHARED_PTR_IMPL_DEF<TrackPredicate> TrackPredicatePtr;
typedef QTPOD_SHARED_PTR_IMPL_DEF<const TrackPredicate> ConstTrackPredicatePtr;


/**
 * This class provides a few TrackPredicate implementations.
 */
class TrackPredicates {
public:

    /**
     * A predicate to find tracks by a given artist. This predicate will return true for
     * every track that fits to the artist given at construction time.
     * @see ByArtist::ByArtist( const QString& )
     */
    class ByArtist : public TrackPredicate {
    public:
        /**
         * Constructor to create a new predicate for the given artist.
         */
        ByArtist( const QString& artist ) : _artist_( artist ) {}
        virtual ~ByArtist() {}

        /**
         * Returns true if the given track matches the parameters given to the constructor.
         */
        bool operator () (const Track& track) {
            return  _artist_.compare( track.getArtist() ) == 0;
        }
    private:
        const QString _artist_;
    };  // ByArtist


    /**
     * A predicate to find tracks by a given album. This predicate will return true for
     * every track that fits to the artist/album combination given at construction time.
     * @see ByAlbum::ByAlbum( const QString&, const QString& )
     */
    class ByAlbum : public TrackPredicate {
    public:
        /**
         * Constructor to create a new predicate for the given artist/album combination.
         */
        ByAlbum(const QString& artist, const QString & album)
            : _artist_( artist ), _album_( album ) {}
        virtual ~ByAlbum() {}

        /**
         * Returns true if the given track matches the parameters given to the constructor.
         */
        bool operator () (const Track& track) {
            bool result = true;
            if ( !_artist_.isEmpty() ) {
                result = _artist_.compare( track.getArtist() ) == 0;
            }
            return  result && _album_.compare( track.getAlbum() ) == 0;
        }
    private:
        const QString _artist_;
        const QString _album_;
    };


    /**
     * A predicate to find the track by a given artist/album/title. This predicate will return
     * true for every track that fits to the artist/album/title combination given at construction
     * time.
     * @see ByFullInfo::ByFullInfo( const QString&, const QString&, const QString&, Q_UINT32 )
     */
    class ByFullInfo : public TrackPredicate {
    public:
        /**
         * Constructor to create a new predicate for the given artist/album/title combination.
         */
        ByFullInfo( const QString& artist, const QString& album, const QString& title, Q_UINT32 trackNum = 0 )
            : m_artist( artist ), m_album( album ), m_title( title ), m_trackNum( trackNum ) {}
        virtual ~ByFullInfo() {}

        bool operator () (const Track& track) {
            return  m_artist.compare( track.getArtist() ) == 0 &&
                    m_album.compare( track.getAlbum() ) == 0 &&
                    m_title.compare( track.getTitle() ) == 0 &&
                    ( m_trackNum == track.getTrackNumber() || m_trackNum == 0 );
        }
    private:
        const QString& m_artist;
        const QString& m_album;
        const QString& m_title;
        Q_UINT32 m_trackNum;
    };  // ByFullInfo


    /**
     * Predicate to find tracks by a given free text search. This predicate will return true
     * for every track that somehow contains the string given at construction time.
     * @see Contains::Contains( const QString&, bool )
     */
    class Contains : public TrackPredicate {
    public:
        /**
         * Constructor to create a new predicate for the given free text string.
         */
        Contains( const QString& someString, bool caseSensitive )
            : m_someString( someString ), m_caseSensitive( caseSensitive ) {}
        virtual ~Contains() {}

        /**
         * Returns true if the given track somehow contains the string given at construction time.
         */
        bool operator () ( const Track& track ) {
            QString sBuffer( track.getArtist() + track.getAlbum() + track.getTitle() + track.getComposer() );
            return sBuffer.contains( m_someString, m_caseSensitive );
        }
    private:
        const QString m_someString;
        bool m_caseSensitive;
    };  // Contains


    class TrackPtrPredicate {
        TrackPredicate & delegate;
    public:
        TrackPtrPredicate( TrackPredicate& predicate ) : delegate( predicate ) {}
        ~TrackPtrPredicate() {}
        bool operator() ( const Track * track ) {
            if ( !track ) {
                return false;
            }
            return delegate( *track );
        }
    };
};  // TrackPredicates


/**
 * A List implementation for Track pointers.
 */
class TrackPtrList : public QList<itunesdb::Track> {
    TrackComparatorPtr m_comparator;
public:

    /**
     * Default constructor.
     */
    TrackPtrList();
    virtual ~TrackPtrList();

    /**
     * This method sets the comparator that compares two tracks for sorting.
     * TrackPtrList owns the given pointer so it may be deallocated at any time.
     * @param comparator the comparator to set. If NULL is given the DefaultComparator will be used.
     */
    void setComparator( const TrackComparatorPtr& comparator );

    /**
     * Returns the currently used comparator for sorting.
     */
    const TrackComparatorPtr& getComparator() const;

    /**
     * This method sets the comparator that compares two tracks for sorting.
     * TrackPtrList owns the given pointer so it may be deallocated at any time.
     * @param comparator the comparator to set. If NULL is given the DefaultComparator will be used.
     */
    void setComparator( TrackComparator * comparator );

    /**
     * Compares two tracks by the comparator set by @c setComparator().
     */
    virtual int compareItems( QPtrCollection::Item s1, QPtrCollection::Item s2 );

    /**
     * Returns true if this list contains tracks not in sync with the database.
     */
    bool hasDirtyTracks();

    /**
     * Returns true if this list contains tracks being played after the last
     * sync.
     */
    bool hasRecentlyPlayedTracks();

    /**
     * removes all references to the given track.
     */
    bool removeAllRefs( itunesdb::Track * track );

    /**
     * Returns all the Tracks where the given predicate returned true
     */
    TrackPtrList * getTracksBy( TrackPredicate& predicate, itunesdb::TrackPtrList& buffer ) const;

    /**
     * Returns the first Track where the given predicate returned true
     * The given trackpredicate needs to contain a method that gets a track, and returns a bool
     * if it's the desired track
     * @return the first Track where the given predicate returned true
     * @see class TrackPredicate
     */
    Track * findFirstTrackBy( TrackPredicate& predicate ) const;

    /**
     * Adds all tracks delivered by the given Iterator.
     */
    template <class JavaLikeTrackIterator>
        void addAll( JavaLikeTrackIterator trackIter ) {
            while ( trackIter.hasNext() ) {
                append( trackIter.next() );
            }
        }
};

}   // namespace itunesdb

#endif
