 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#ifndef _qtpod_itunesdb_utils_H_
#define _qtpod_itunesdb_utils_H_

// here you can enable usage of the boost shared_ptr impl if using an older libstdc++
// uncomment the following line to use the boost shared_ptr
// #define QTPOD_USE_BOOST_SHAREDPTR 1

#ifdef QTPOD_USE_BOOST_SHAREDPTR

#include <boost/shared_ptr.hpp>
#define QTPOD_SHARED_PTR_IMPL_DEF boost::shared_ptr

#else

#include <tr1/memory>
#define QTPOD_SHARED_PTR_IMPL_DEF std::tr1::shared_ptr

#endif

#define MAC_EPOCH_DELTA 2082844800

namespace itunesdb {

namespace utils {

/**
 * some stuff to find/filter etc through containers
 */

/**
 * Returns the Iterator where predicate(*iter) == TRUE
 */
template <class Iter, class TUnaryPredicate>
Iter findFirst( Iter pos, Iter end, const TUnaryPredicate& predicate ) {
    for (; pos != end; ++pos) {
        if (predicate(*pos)) {
            return pos;
        }
    }
    return end;
}

/**
 * Returns the Iterator to the element where the given function returned the biggest result
 */
template <class Iter, class TUnaryFunction>
Iter max( Iter pos, Iter end, const TUnaryFunction& fn ) {
    Iter currentMax = pos;
    for ( ; pos != end; ++pos ) {
        if ( fn( *pos ) > fn ( *currentMax ) ) {
            currentMax = pos;
        }
    }
    return currentMax;
}

template <class ContainerT, class Iter>
ContainerT& appendAll( ContainerT& container, Iter begin, Iter end ) {
    Iter elemIter = begin;
    for ( ; elemIter != end; ++elemIter ) {
        container.append( *elemIter );
    }
    return container;
}

/**
 * An Iterator over those elements where the given predicate returned TRUE
 */
template <class Iter, class TUnaryPredicate>
class FilteredIterator {
public:
    /**
     * Constructs a new Iterator from start to end with the given predicate.
     */
    FilteredIterator(Iter start, Iter end, TUnaryPredicate predicate)
        : _pos_(start), _end_(end), _next_(NULL), _unarypredicate_(predicate) {}

    /**
     * Returns true if there are elements left.
     */
    bool hasNext() {
        calcNext();
        return _next_ != _end_;
    }

    /**
     * Returns the next element of the collection of elements we iterate over where the
     * given predicate returned true.
     */
    Iter next() {
        calcNext();
        Iter result = _next_;
        _next_ = _end_;
        ++_pos_;
        return result;
    }
private:
    void calcNext() {
        while ( _pos_ != _end_ && _next_ == _end_ ) {
            if (_unarypredicate_(*_pos_)) {
                _next_ = _pos_;
            } else {
                ++_pos_;
            }
        }
    }
    Iter _pos_;
    Iter _end_;
    Iter _next_;
    TUnaryPredicate _unarypredicate_;
};

/**
 * inherit from this to get a non copyable class
 */
class NonCopyAble {
private:
    NonCopyAble( const NonCopyAble& ) {}
protected:
    NonCopyAble() {}
public:
    virtual ~NonCopyAble() {}
};


/**
 * Seeks the given number of bytes in the given stream
 */
template <class Stream_T>
inline void seekRelative( Stream_T& stream, uint numbytes ) {
    if ( numbytes > 0 ) {
        char * buffer = new char[ numbytes ];
        stream.readRawBytes( buffer, numbytes );
        delete [] buffer;
    }
}


} // utils

} // itunesdb

#endif
